"""Tests for GPU discovery and management utilities."""

import os
from unittest.mock import patch

import pytest


def test_discover_gpus_from_slurm_job_gpus(monkeypatch):
    """Test GPU discovery from SLURM_JOB_GPUS environment variable."""
    from genome_entropy.encode3di.gpu_utils import discover_available_gpus
    
    # Set SLURM_JOB_GPUS
    monkeypatch.setenv("SLURM_JOB_GPUS", "0,1,2")
    monkeypatch.delenv("SLURM_GPUS", raising=False)
    monkeypatch.delenv("CUDA_VISIBLE_DEVICES", raising=False)
    
    gpu_ids = discover_available_gpus()
    assert gpu_ids == [0, 1, 2]


def test_discover_gpus_from_slurm_gpus(monkeypatch):
    """Test GPU discovery from SLURM_GPUS environment variable."""
    from genome_entropy.encode3di.gpu_utils import discover_available_gpus
    
    # Set SLURM_GPUS (when SLURM_JOB_GPUS is not set)
    monkeypatch.delenv("SLURM_JOB_GPUS", raising=False)
    monkeypatch.setenv("SLURM_GPUS", "2,3,4")
    monkeypatch.delenv("CUDA_VISIBLE_DEVICES", raising=False)
    
    gpu_ids = discover_available_gpus()
    assert gpu_ids == [2, 3, 4]


def test_discover_gpus_from_cuda_visible_devices(monkeypatch):
    """Test GPU discovery from CUDA_VISIBLE_DEVICES."""
    from genome_entropy.encode3di.gpu_utils import discover_available_gpus
    
    # Set CUDA_VISIBLE_DEVICES (when SLURM vars are not set)
    monkeypatch.delenv("SLURM_JOB_GPUS", raising=False)
    monkeypatch.delenv("SLURM_GPUS", raising=False)
    monkeypatch.setenv("CUDA_VISIBLE_DEVICES", "5,6")
    
    gpu_ids = discover_available_gpus()
    # CUDA_VISIBLE_DEVICES remaps to local indices
    assert gpu_ids == [0, 1]


def test_discover_gpus_priority_slurm_job_over_slurm(monkeypatch):
    """Test that SLURM_JOB_GPUS has priority over SLURM_GPUS."""
    from genome_entropy.encode3di.gpu_utils import discover_available_gpus
    
    # Set both, SLURM_JOB_GPUS should win
    monkeypatch.setenv("SLURM_JOB_GPUS", "0,1")
    monkeypatch.setenv("SLURM_GPUS", "2,3,4")
    monkeypatch.delenv("CUDA_VISIBLE_DEVICES", raising=False)
    
    gpu_ids = discover_available_gpus()
    assert gpu_ids == [0, 1]  # From SLURM_JOB_GPUS


def test_discover_gpus_priority_slurm_over_cuda(monkeypatch):
    """Test that SLURM vars have priority over CUDA_VISIBLE_DEVICES."""
    from genome_entropy.encode3di.gpu_utils import discover_available_gpus
    
    # Set SLURM_GPUS and CUDA_VISIBLE_DEVICES, SLURM should win
    monkeypatch.setenv("SLURM_GPUS", "1,2")
    monkeypatch.setenv("CUDA_VISIBLE_DEVICES", "5,6,7")
    monkeypatch.delenv("SLURM_JOB_GPUS", raising=False)
    
    gpu_ids = discover_available_gpus()
    assert gpu_ids == [1, 2]  # From SLURM_GPUS


def test_discover_gpus_fallback_to_torch_cuda(monkeypatch):
    """Test fallback to torch.cuda when no env vars are set."""
    from genome_entropy.encode3di.gpu_utils import discover_available_gpus
    
    # Clear all env vars
    monkeypatch.delenv("SLURM_JOB_GPUS", raising=False)
    monkeypatch.delenv("SLURM_GPUS", raising=False)
    monkeypatch.delenv("CUDA_VISIBLE_DEVICES", raising=False)
    
    # Mock torch.cuda
    with patch("genome_entropy.encode3di.gpu_utils.torch") as mock_torch:
        mock_torch.cuda.is_available.return_value = True
        mock_torch.cuda.device_count.return_value = 4
        
        gpu_ids = discover_available_gpus()
        assert gpu_ids == [0, 1, 2, 3]


def test_discover_gpus_no_gpus_available(monkeypatch):
    """Test when no GPUs are available."""
    from genome_entropy.encode3di.gpu_utils import discover_available_gpus
    
    # Clear all env vars
    monkeypatch.delenv("SLURM_JOB_GPUS", raising=False)
    monkeypatch.delenv("SLURM_GPUS", raising=False)
    monkeypatch.delenv("CUDA_VISIBLE_DEVICES", raising=False)
    
    # Mock torch.cuda as unavailable
    with patch("genome_entropy.encode3di.gpu_utils.torch") as mock_torch:
        mock_torch.cuda.is_available.return_value = False
        
        gpu_ids = discover_available_gpus()
        assert gpu_ids == []


def test_discover_gpus_invalid_slurm_format(monkeypatch, caplog):
    """Test handling of invalid SLURM_JOB_GPUS format."""
    from genome_entropy.encode3di.gpu_utils import discover_available_gpus
    
    # Set invalid format
    monkeypatch.setenv("SLURM_JOB_GPUS", "invalid,format,x")
    monkeypatch.delenv("SLURM_GPUS", raising=False)
    monkeypatch.delenv("CUDA_VISIBLE_DEVICES", raising=False)
    
    # Mock torch.cuda as unavailable (so it doesn't fallback)
    with patch("genome_entropy.encode3di.gpu_utils.torch") as mock_torch:
        mock_torch.cuda.is_available.return_value = False
        
        gpu_ids = discover_available_gpus()
        assert gpu_ids == []  # Falls back to empty
        assert "Failed to parse SLURM_JOB_GPUS" in caplog.text


def test_discover_gpus_empty_string(monkeypatch):
    """Test handling of empty GPU string."""
    from genome_entropy.encode3di.gpu_utils import discover_available_gpus
    
    # Set empty string
    monkeypatch.setenv("SLURM_JOB_GPUS", "")
    monkeypatch.delenv("SLURM_GPUS", raising=False)
    monkeypatch.delenv("CUDA_VISIBLE_DEVICES", raising=False)
    
    # Mock torch.cuda as unavailable
    with patch("genome_entropy.encode3di.gpu_utils.torch") as mock_torch:
        mock_torch.cuda.is_available.return_value = False
        
        gpu_ids = discover_available_gpus()
        assert gpu_ids == []


def test_parse_gpu_list():
    """Test GPU list parsing."""
    from genome_entropy.encode3di.gpu_utils import _parse_gpu_list
    
    assert _parse_gpu_list("0,1,2") == [0, 1, 2]
    assert _parse_gpu_list("5") == [5]
    assert _parse_gpu_list("0, 1, 2") == [0, 1, 2]  # With spaces
    assert _parse_gpu_list("") == []
    assert _parse_gpu_list("  ") == []


def test_parse_gpu_list_invalid():
    """Test GPU list parsing with invalid input."""
    from genome_entropy.encode3di.gpu_utils import _parse_gpu_list
    
    with pytest.raises(ValueError):
        _parse_gpu_list("a,b,c")
    
    with pytest.raises(ValueError):
        _parse_gpu_list("0,1,invalid")


def test_select_device_for_gpu():
    """Test device string generation."""
    from genome_entropy.encode3di.gpu_utils import select_device_for_gpu
    
    assert select_device_for_gpu(0) == "cuda:0"
    assert select_device_for_gpu(1) == "cuda:1"
    assert select_device_for_gpu(5) == "cuda:5"


def test_validate_gpu_availability():
    """Test GPU availability validation."""
    from genome_entropy.encode3di.gpu_utils import validate_gpu_availability
    
    # Mock torch.cuda with 4 devices
    with patch("genome_entropy.encode3di.gpu_utils.torch") as mock_torch:
        mock_torch.cuda.is_available.return_value = True
        mock_torch.cuda.device_count.return_value = 4
        
        # Valid IDs
        assert validate_gpu_availability([0, 1, 2]) == [0, 1, 2]
        assert validate_gpu_availability([0]) == [0]
        assert validate_gpu_availability([3]) == [3]
        
        # Invalid IDs (out of range)
        assert validate_gpu_availability([4, 5]) == []
        assert validate_gpu_availability([0, 1, 10]) == [0, 1]
        
        # Empty input
        assert validate_gpu_availability([]) == []


def test_validate_gpu_availability_no_cuda():
    """Test GPU validation when CUDA is not available."""
    from genome_entropy.encode3di.gpu_utils import validate_gpu_availability
    
    # Mock torch.cuda as unavailable
    with patch("genome_entropy.encode3di.gpu_utils.torch") as mock_torch:
        mock_torch.cuda.is_available.return_value = False
        
        assert validate_gpu_availability([0, 1, 2]) == []


def test_validate_gpu_availability_no_torch(monkeypatch):
    """Test GPU validation when torch is not available."""
    from genome_entropy.encode3di import gpu_utils
    from genome_entropy.encode3di.gpu_utils import validate_gpu_availability
    
    # Mock torch as None
    monkeypatch.setattr(gpu_utils, "torch", None)
    
    assert validate_gpu_availability([0, 1, 2]) == []



"""Basic tests to verify package installation and structure."""

import genome_entropy


def test_version():
    """Test that version is defined."""
    assert hasattr(genome_entropy, "__version__")
    assert isinstance(genome_entropy.__version__, str)


def test_package_import():
    """Test that the package can be imported."""
    assert genome_entropy is not None



"""Token size estimation command."""

from pathlib import Path
from typing import Optional
import traceback

try:
    import typer
except ImportError:
    typer = None


def estimate_token_size_command(
    model: str = typer.Option(
        "Rostlab/ProstT5_fp16",
        "--model",
        "-m",
        help="ProstT5 model name",
    ),
    device: Optional[str] = typer.Option(
        None,
        "--device",
        "-d",
        help="Device to use (auto/cuda/mps/cpu)",
    ),
    start_length: int = typer.Option(
        3000,
        "--start",
        "-s",
        help="Starting total protein length to test (amino acids)",
    ),
    end_length: int = typer.Option(
        10000,
        "--end",
        "-e",
        help="Maximum total protein length to test (amino acids)",
    ),
    step: int = typer.Option(
        1000,
        "--step",
        help="Increment between test lengths (amino acids)",
    ),
    num_trials: int = typer.Option(
        3,
        "--trials",
        "-t",
        help="Number of trials per length for robustness",
    ),
    base_protein_length: int = typer.Option(
        100,
        "--base-length",
        "-b",
        help="Approximate length of individual proteins (amino acids)",
    ),
    log_level: str = typer.Option(
        "INFO",
        "--log-level",
        "-l",
        help="Logging level (DEBUG, INFO, WARNING, ERROR, CRITICAL)",
    ),
    log_file: Optional[Path] = typer.Option(
        None,
        "--log-file",
        help="Path to log file (default: log to STDOUT)",
    ),
) -> None:
    """Estimate optimal token size for GPU encoding.

    This command generates random protein sequences of increasing length
    and tests encoding to find the maximum that can be encoded on the
    available GPU without running out of memory.

    The recommended token size is returned as 90% of the maximum for safety.
    """
    try:
        from ...config import VALID_LOG_LEVELS
        from ...logging_config import configure_logging
        from ...encode3di import ProstT5ThreeDiEncoder, estimate_token_size

        # Validate and configure logging
        if log_level.upper() not in VALID_LOG_LEVELS:
            typer.echo(
                f"Error: Invalid log level '{log_level}'. Must be one of: {', '.join(VALID_LOG_LEVELS)}",
                err=True,
            )
            raise typer.Exit(2)

        configure_logging(level=log_level.upper(), log_file=log_file)

        typer.echo("=" * 60)
        typer.echo("Token Size Estimation")
        typer.echo("=" * 60)
        typer.echo(f"Model: {model}")
        typer.echo(f"Testing range: {start_length} to {end_length} AA (step: {step})")
        typer.echo(f"Trials per length: {num_trials}")
        typer.echo(f"Base protein length: {base_protein_length} AA")
        typer.echo("=" * 60)

        typer.echo("\nInitializing ProstT5 encoder...")
        encoder = ProstT5ThreeDiEncoder(model_name=model, device=device)
        typer.echo(f"Using device: {encoder.device}")

        typer.echo("\nStarting estimation (this may take several minutes)...")
        typer.echo("Watch for Out of Memory errors to identify the limit.\n")

        results = estimate_token_size(
            encoder=encoder,
            start_length=start_length,
            end_length=end_length,
            step=step,
            num_trials=num_trials,
            base_protein_length=base_protein_length,
        )

        typer.echo("\n" + "=" * 60)
        typer.echo("RESULTS")
        typer.echo("=" * 60)
        typer.echo(f"Device: {results['device']}")
        typer.echo(f"Max successful length: {results['max_length']} amino acids")
        typer.echo(
            f"Recommended token size: {results['recommended_token_size']} amino acids"
        )
        typer.echo("\nTrials per length:")
        for length, trials in sorted(results["trials_per_length"].items()):
            status = "✓" if trials > 0 else "✗"
            typer.echo(f"  {status} {length} AA: {trials}/{num_trials} successful")
        typer.echo("=" * 60)
        typer.echo("\n✓ Estimation complete!")
        typer.echo(
            f"\nTo use this in encoding, add: --encoding-size {results['recommended_token_size']}"
        )

    except Exception as e:
        typer.echo(f"Error: {e}", err=True)
        traceback.print_exc()
        raise typer.Exit(3)



"""ORF finder wrapper using get_orfs binary."""

import subprocess
import tempfile
import re
import sys
from pathlib import Path
from typing import Dict, List, Literal

from ..config import DEFAULT_GENETIC_CODE_TABLE, GET_ORFS_BINARY
from ..errors import OrfFinderError
from ..logging_config import get_logger
from .types import OrfRecord

logger = get_logger(__name__)


def find_orfs(
    sequences: Dict[str, str],
    table_id: int = DEFAULT_GENETIC_CODE_TABLE,
    min_nt_length: int = 90,
    binary_path: str = GET_ORFS_BINARY,
) -> List[OrfRecord]:
    """Find ORFs in DNA sequences using get_orfs binary.

    This function wraps the external get_orfs binary (https://github.com/linsalrob/get_orfs).
    The binary must be installed and available in PATH or specified via binary_path.

    Args:
        sequences: Dictionary mapping sequence IDs to DNA sequences
        table_id: NCBI genetic code table ID (default: 11, bacterial)
        min_nt_length: Minimum ORF length in nucleotides (default: 90)
        binary_path: Path to get_orfs binary (default: from config/environment)

    Returns:
        List of OrfRecord objects

    Raises:
        OrfFinderError: If get_orfs binary is not found or fails
    """
    logger.info(
        "Starting ORF finding for %d sequence(s) (table=%d, min_length=%d)",
        len(sequences),
        table_id,
        min_nt_length,
    )

    # Check if binary exists
    try:
        logger.debug("Checking get_orfs binary at: %s", binary_path)
        result = subprocess.run(
            [binary_path, "-v"],
            capture_output=True,
            text=True,
            timeout=5,
        )
        if result.returncode != 0:
            raise OrfFinderError(f"get_orfs binary check failed: {result.stderr}")
        logger.debug("get_orfs binary check successful")
    except FileNotFoundError:
        logger.error("get_orfs binary not found at: %s", binary_path)
        raise OrfFinderError(
            f"get_orfs binary not found at: {binary_path}\n"
            "Please install from https://github.com/linsalrob/get_orfs\n"
            "or set GET_ORFS_PATH environment variable."
        )
    except subprocess.TimeoutExpired:
        logger.error("get_orfs binary check timed out")
        raise OrfFinderError("get_orfs binary check timed out")

    # Write sequences to temporary FASTA file
    with tempfile.NamedTemporaryFile(
        mode="w", suffix=".fasta", delete=False
    ) as tmp_fasta:
        tmp_fasta_path = tmp_fasta.name
        for seq_id, sequence in sequences.items():
            tmp_fasta.write(f">{seq_id}\n{sequence}\n")

    logger.debug("Wrote sequences to temporary file: %s", tmp_fasta_path)

    try:
        # Run get_orfs
        # Expected command: get_orfs -f fasta_file -t table_id -l min_length
        cmd = [
            binary_path,
            "-f",
            tmp_fasta_path,
            "-t",
            str(table_id),
            "-l",
            str(min_nt_length),
        ]

        logger.debug("Running command: %s", " ".join(cmd))
        result = subprocess.run(
            cmd,
            capture_output=True,
            text=True,
            timeout=300,  # 5 minute timeout
        )

        if result.returncode != 0:
            logger.error(
                "get_orfs failed with return code %d: %s",
                result.returncode,
                result.stderr,
            )
            raise OrfFinderError(f"get_orfs failed: {result.stderr}")

        logger.debug("get_orfs completed successfully")

        # Parse output
        orfs = _parse_get_orfs_output(result.stdout, sequences, table_id)
        logger.info("Found %d ORF(s) in %d sequence(s)", len(orfs), len(sequences))
        return orfs

    finally:
        # Clean up temporary file
        Path(tmp_fasta_path).unlink(missing_ok=True)
        logger.debug("Cleaned up temporary file: %s", tmp_fasta_path)


def _parse_orf_header_line(header: str, table_id: int) -> OrfRecord:
    """Parse ORF header line from get_orfs output.

    Supports two formats:
    1. Standard: >parent_id-orf_id [parent_id frame frame_num start end]
    2. Simple:   >orf_id [parent_id frame frame_num start end]

    Args:
        header: FASTA header line starting with >
        table_id: Genetic code table ID

    Returns:
        OrfRecord object

    Raises:
        ValueError: If header format is invalid
    """
    # Pattern for standard format: >parent_id-orf_id [parent_id frame ...]
    FASTA_RE_STANDARD = re.compile(
        r"""
        ^>
        (?P<parent_id>[^-\s]+)      # JQ995537
        -
        (?P<orf_id>orf\d+)          # orf14635
        \s+
        \[
            (?P=parent_id)          # repeat accession
            \s+frame\s+
            (?P<frame>[+-]?\d+)
            \s+
            (?P<start>\d+)
            \s+
            (?P<end>\d+)
        \]
        $
    """,
        re.VERBOSE,
    )

    # Pattern for simple format: >orf_id [parent_id frame ...]
    FASTA_RE_SIMPLE = re.compile(
        r"""
        ^>
        (?P<orf_id>orf\d+)          # orf1
        \s+
        \[
            (?P<parent_id>[^\s]+)   # JQ995537
            \s+frame\s+
            (?P<frame>[+-]?\d+)
            \s+
            (?P<start>\d+)
            \s+
            (?P<end>\d+)
        \]
        $
    """,
        re.VERBOSE,
    )

    # Try standard format first
    m = FASTA_RE_STANDARD.match(header)
    if not m:
        # Try simple format
        m = FASTA_RE_SIMPLE.match(header)

    if not m:
        raise ValueError("Header did not match expected format")

    data = m.groupdict()
    strand: Literal["+", "-"] = "+"
    if int(data["frame"]) < 0:
        strand = "-"

    orf = OrfRecord(
        parent_id=data["parent_id"],
        orf_id=data["orf_id"],
        start=int(data["start"]),
        end=int(data["end"]),
        frame=abs(int(data["frame"])),
        strand=strand,
        nt_sequence="",
        aa_sequence="",
        table_id=table_id,
        has_start_codon=False,
        has_stop_codon=False,
    )

    return orf


def reverse_complement(seq: str) -> str:
    """Return the reverse complement of a DNA sequence."""
    complement = str.maketrans("ACGTacgt", "TGCAtgca")
    return seq.translate(complement)[::-1]


def _extract_orf_dna_sequence(sequences: Dict[str, str], current_orf: OrfRecord) -> str:
    """
    Extract the sequence from sequences
    """

    if current_orf.parent_id not in sequences:
        raise ValueError(
            f"We were asked to extract an orf for {current_orf.parent_id} but it is not in our sequences: {sequences.keys()}"
        )

    if current_orf.start < 1 or current_orf.end < 1:
        raise ValueError("Coordinates must be >= 1")

    if current_orf.end < current_orf.start:
        raise ValueError("End coordinate precedes start coordinate")

    # Convert to Python slicing (0-based, end-exclusive)
    dna = sequences[current_orf.parent_id]
    if current_orf.strand == "-":
        dna = reverse_complement(dna)

    orf_seq = dna[current_orf.start - 1 : current_orf.end]

    return orf_seq


def _parse_get_orfs_output(
    output: str, sequences: Dict[str, str], table_id: int
) -> List[OrfRecord]:
    """Parse get_orfs output into OrfRecord objects.

    Format: >JQ995537-orf14635 [JQ995537 frame -3 96951 97093]

    Args:
        output: stdout from get_orfs
        sequences: Original DNA sequences (for validation)
        table_id: Genetic code table used

    Returns:
        List of OrfRecord objects
    """
    logger.debug("Parsing get_orfs output")
    orfs = []
    current_orf = None

    for line in output.strip().split("\n"):
        try:
            line = line.rstrip("\r\n")
            if line.startswith(">"):
                if current_orf is not None:
                    current_orf.nt_sequence = _extract_orf_dna_sequence(
                        sequences, current_orf
                    )
                    current_orf.has_start_codon = (
                        True if "M" in current_orf.aa_sequence else False
                    )
                    current_orf.has_stop_codon = (
                        True if "*" in current_orf.aa_sequence else False
                    )
                    orfs.append(current_orf)
                    logger.debug(
                        "Parsed ORF: %s (length=%d nt)",
                        current_orf.orf_id,
                        len(current_orf.nt_sequence),
                    )
                current_orf = _parse_orf_header_line(line, table_id)
            elif current_orf is not None:
                current_orf.aa_sequence += line.strip()
        except (ValueError, IndexError) as e:
            logger.error("Error handling sequence line '%s': %s", line, e)
            print(f"Error handling sequence {line}", file=sys.stderr)
            raise Exception(f"Error as {e}")

    # Add the last ORF if present
    if current_orf is not None:
        current_orf.nt_sequence = _extract_orf_dna_sequence(sequences, current_orf)
        current_orf.has_start_codon = True if "M" in current_orf.aa_sequence else False
        current_orf.has_stop_codon = True if "*" in current_orf.aa_sequence else False
        orfs.append(current_orf)
        logger.debug(
            "Parsed ORF: %s (length=%d nt)",
            current_orf.orf_id,
            len(current_orf.nt_sequence),
        )

    logger.debug("Parsed %d ORF(s) from get_orfs output", len(orfs))
    return orfs



"""Tests for multi-GPU asynchronous encoding."""

import asyncio
from unittest.mock import MagicMock, patch

import pytest


def test_multi_gpu_encoder_initialization():
    """Test MultiGPUEncoder initialization with specified GPUs."""
    from genome_entropy.encode3di.multi_gpu import MultiGPUEncoder
    
    # Mock encoder class
    mock_encoder_class = MagicMock()
    mock_encoder_instance = MagicMock()
    mock_encoder_class.return_value = mock_encoder_instance
    
    # Test with specified GPU IDs
    with patch("genome_entropy.encode3di.multi_gpu.validate_gpu_availability") as mock_validate:
        mock_validate.return_value = [0, 1, 2]
        
        encoder = MultiGPUEncoder(
            model_name="test-model",
            encoder_class=mock_encoder_class,
            gpu_ids=[0, 1, 2],
        )
        
        assert encoder.num_gpus == 3
        assert encoder.is_multi_gpu()
        assert len(encoder.encoders) == 3


def test_multi_gpu_encoder_auto_discovery():
    """Test MultiGPUEncoder with automatic GPU discovery."""
    from genome_entropy.encode3di.multi_gpu import MultiGPUEncoder
    
    # Mock encoder class
    mock_encoder_class = MagicMock()
    
    # Mock GPU discovery
    with patch("genome_entropy.encode3di.multi_gpu.discover_available_gpus") as mock_discover:
        with patch("genome_entropy.encode3di.multi_gpu.validate_gpu_availability") as mock_validate:
            mock_discover.return_value = [0, 1]
            mock_validate.return_value = [0, 1]
            
            encoder = MultiGPUEncoder(
                model_name="test-model",
                encoder_class=mock_encoder_class,
                gpu_ids=None,  # Auto-discover
            )
            
            assert encoder.num_gpus == 2
            assert encoder.is_multi_gpu()


def test_multi_gpu_encoder_fallback_no_gpus():
    """Test MultiGPUEncoder fallback when no GPUs are available."""
    from genome_entropy.encode3di.multi_gpu import MultiGPUEncoder
    
    # Mock encoder class
    mock_encoder_class = MagicMock()
    
    # Mock no GPUs available
    with patch("genome_entropy.encode3di.multi_gpu.discover_available_gpus") as mock_discover:
        mock_discover.return_value = []
        
        encoder = MultiGPUEncoder(
            model_name="test-model",
            encoder_class=mock_encoder_class,
            gpu_ids=None,
        )
        
        # Should fall back to single encoder
        assert encoder.num_gpus == 1
        assert not encoder.is_multi_gpu()


@pytest.mark.asyncio
async def test_encode_batch_async():
    """Test asynchronous batch encoding."""
    from genome_entropy.encode3di.multi_gpu import MultiGPUEncoder
    from genome_entropy.encode3di.types import IndexedSeq
    
    # Mock encoder class and instance
    mock_encoder_class = MagicMock()
    mock_encoder_instance = MagicMock()
    mock_encoder_instance._encode_batch = MagicMock(return_value=["aaa", "bbb"])
    mock_encoder_class.return_value = mock_encoder_instance
    
    with patch("genome_entropy.encode3di.multi_gpu.validate_gpu_availability") as mock_validate:
        mock_validate.return_value = [0]
        
        encoder = MultiGPUEncoder(
            model_name="test-model",
            encoder_class=mock_encoder_class,
            gpu_ids=[0],
        )
        
        # Create test batch
        batch = [
            IndexedSeq(idx=0, seq="AAA"),
            IndexedSeq(idx=1, seq="BBB"),
        ]
        
        # Encode batch
        indices, results = await encoder.encode_batch_async(0, batch)
        
        assert indices == [0, 1]
        assert results == ["aaa", "bbb"]


@pytest.mark.asyncio
async def test_encode_all_batches_async():
    """Test encoding multiple batches across GPUs."""
    from genome_entropy.encode3di.multi_gpu import MultiGPUEncoder
    from genome_entropy.encode3di.types import IndexedSeq
    
    # Mock encoder class and instances
    mock_encoder_class = MagicMock()
    mock_encoder1 = MagicMock()
    mock_encoder2 = MagicMock()
    
    # Mock encoding results
    mock_encoder1._encode_batch = MagicMock(return_value=["aaa", "bbb"])
    mock_encoder2._encode_batch = MagicMock(return_value=["ccc"])
    
    mock_encoder_class.side_effect = [mock_encoder1, mock_encoder2]
    
    with patch("genome_entropy.encode3di.multi_gpu.validate_gpu_availability") as mock_validate:
        mock_validate.return_value = [0, 1]
        
        encoder = MultiGPUEncoder(
            model_name="test-model",
            encoder_class=mock_encoder_class,
            gpu_ids=[0, 1],
        )
        
        # Create test batches
        batches = [
            [IndexedSeq(idx=0, seq="AAA"), IndexedSeq(idx=1, seq="BBB")],  # Batch 0 -> GPU 0
            [IndexedSeq(idx=2, seq="CCC")],  # Batch 1 -> GPU 1
        ]
        
        # Encode all batches
        results = await encoder.encode_all_batches_async(batches, total_sequences=3)
        
        # Check results are in correct order
        assert len(results) == 3
        assert results[0] == "aaa"
        assert results[1] == "bbb"
        assert results[2] == "ccc"


def test_encode_multi_gpu_with_multiple_gpus():
    """Test multi-GPU encoding with multiple GPUs."""
    from genome_entropy.encode3di.multi_gpu import MultiGPUEncoder
    from genome_entropy.encode3di.types import IndexedSeq
    
    # Mock encoder class and instances
    mock_encoder_class = MagicMock()
    mock_encoder1 = MagicMock()
    mock_encoder2 = MagicMock()
    
    mock_encoder1._encode_batch = MagicMock(return_value=["aaa"])
    mock_encoder1._load_model = MagicMock()
    mock_encoder2._encode_batch = MagicMock(return_value=["bbb"])
    mock_encoder2._load_model = MagicMock()
    
    mock_encoder_class.side_effect = [mock_encoder1, mock_encoder2]
    
    # Mock token budget batches function
    def mock_batches(seqs, budget):
        return [
            [IndexedSeq(idx=0, seq=seqs[0])],
            [IndexedSeq(idx=1, seq=seqs[1])],
        ]
    
    with patch("genome_entropy.encode3di.multi_gpu.validate_gpu_availability") as mock_validate:
        mock_validate.return_value = [0, 1]
        
        encoder = MultiGPUEncoder(
            model_name="test-model",
            encoder_class=mock_encoder_class,
            gpu_ids=[0, 1],
        )
        
        # Encode sequences
        results = encoder.encode_multi_gpu(
            aa_sequences=["AAA", "BBB"],
            token_budget_batches_fn=mock_batches,
            encoding_size=10,
        )
        
        assert len(results) == 2
        assert results == ["aaa", "bbb"]


def test_encode_multi_gpu_single_gpu_fallback():
    """Test that single GPU falls back to sequential processing."""
    from genome_entropy.encode3di.multi_gpu import MultiGPUEncoder
    from genome_entropy.encode3di.types import IndexedSeq
    
    # Mock encoder class
    mock_encoder_class = MagicMock()
    mock_encoder = MagicMock()
    mock_encoder._encode_batch = MagicMock(side_effect=[["aaa"], ["bbb"]])
    mock_encoder._load_model = MagicMock()
    mock_encoder_class.return_value = mock_encoder
    
    # Mock token budget batches function
    def mock_batches(seqs, budget):
        return [
            [IndexedSeq(idx=0, seq=seqs[0])],
            [IndexedSeq(idx=1, seq=seqs[1])],
        ]
    
    with patch("genome_entropy.encode3di.multi_gpu.discover_available_gpus") as mock_discover:
        mock_discover.return_value = []
        
        encoder = MultiGPUEncoder(
            model_name="test-model",
            encoder_class=mock_encoder_class,
            gpu_ids=None,
        )
        
        # Should use single GPU
        assert not encoder.is_multi_gpu()
        
        # Encode sequences
        results = encoder.encode_multi_gpu(
            aa_sequences=["AAA", "BBB"],
            token_budget_batches_fn=mock_batches,
            encoding_size=10,
        )
        
        assert len(results) == 2
        assert results == ["aaa", "bbb"]


@pytest.mark.asyncio
async def test_encode_all_batches_error_handling():
    """Test error handling in multi-GPU encoding."""
    from genome_entropy.encode3di.multi_gpu import MultiGPUEncoder
    from genome_entropy.encode3di.types import IndexedSeq
    from genome_entropy.errors import EncodingError
    
    # Mock encoder class
    mock_encoder_class = MagicMock()
    mock_encoder = MagicMock()
    mock_encoder._encode_batch = MagicMock(side_effect=RuntimeError("GPU error"))
    mock_encoder_class.return_value = mock_encoder
    
    with patch("genome_entropy.encode3di.multi_gpu.validate_gpu_availability") as mock_validate:
        mock_validate.return_value = [0]
        
        encoder = MultiGPUEncoder(
            model_name="test-model",
            encoder_class=mock_encoder_class,
            gpu_ids=[0],
        )
        
        batches = [[IndexedSeq(idx=0, seq="AAA")]]
        
        # Should raise EncodingError
        with pytest.raises(EncodingError):
            await encoder.encode_all_batches_async(batches, total_sequences=1)


def test_single_gpu_sequential_encoding():
    """Test sequential encoding for single GPU."""
    from genome_entropy.encode3di.multi_gpu import MultiGPUEncoder
    from genome_entropy.encode3di.types import IndexedSeq
    
    # Mock encoder class
    mock_encoder_class = MagicMock()
    mock_encoder = MagicMock()
    mock_encoder._encode_batch = MagicMock(side_effect=[["aaa"], ["bbb"], ["ccc"]])
    mock_encoder_class.return_value = mock_encoder
    
    with patch("genome_entropy.encode3di.multi_gpu.validate_gpu_availability") as mock_validate:
        mock_validate.return_value = [0]
        
        encoder = MultiGPUEncoder(
            model_name="test-model",
            encoder_class=mock_encoder_class,
            gpu_ids=[0],
        )
        
        batches = [
            [IndexedSeq(idx=0, seq="AAA")],
            [IndexedSeq(idx=1, seq="BBB")],
            [IndexedSeq(idx=2, seq="CCC")],
        ]
        
        results = encoder._encode_single_gpu_sequential(batches, total_sequences=3)
        
        assert len(results) == 3
        assert results == ["aaa", "bbb", "ccc"]



"""Token size estimation for optimal GPU memory usage in 3Di encoding."""

import random
from typing import Any, Dict, List, Optional

try:
    import torch
except ImportError:
    torch = None  # type: ignore[assignment]

from ..config import AA_ALPHABET
from ..logging_config import get_logger

logger = get_logger(__name__)


def generate_random_protein(length: int, seed: Optional[int] = None) -> str:
    """Generate a random protein sequence of specified length.

    Args:
        length: Length of the protein sequence
        seed: Random seed for reproducibility (optional)

    Returns:
        Random protein sequence using the 20 standard amino acids
    """
    if seed is not None:
        random.seed(seed)

    aa_list = list(AA_ALPHABET)
    return "".join(random.choices(aa_list, k=length))


def generate_combined_proteins(
    target_length: int, base_length: int = 100, seed: Optional[int] = None
) -> List[str]:
    """Generate multiple shorter proteins that combine to target length.

    Args:
        target_length: Total target length across all proteins
        base_length: Approximate length of each individual protein
        seed: Random seed for reproducibility (optional)

    Returns:
        List of protein sequences that total approximately target_length
    """
    if seed is not None:
        random.seed(seed)

    proteins = []
    remaining = target_length

    while remaining > 0:
        # Vary protein length slightly for realism
        variation = int(base_length * 0.2)  # 20% variation
        length = random.randint(base_length - variation, base_length + variation)
        length = min(length, remaining)  # Don't exceed target

        proteins.append(generate_random_protein(length))
        remaining -= length

    return proteins


def estimate_token_size(
    encoder: Any,
    start_length: int = 3000,
    end_length: int = 10000,
    step: int = 1000,
    num_trials: int = 3,
    base_protein_length: int = 100,
) -> Dict[str, Any]:
    """Estimate optimal token size for GPU encoding by testing increasing lengths.

    This function generates random protein sequences of increasing total length
    and attempts to encode them. It catches OutOfMemoryError to find the maximum
    length that can be encoded on the available GPU.

    Args:
        encoder: ProstT5ThreeDiEncoder instance to use for encoding
        start_length: Starting total length to test (default: 3000)
        end_length: Maximum total length to test (default: 10000)
        step: Increment between test lengths (default: 1000)
        num_trials: Number of trials per length for robustness (default: 3)
        base_protein_length: Approximate length of individual proteins (default: 100)

    Returns:
        Dictionary with estimation results:
            - 'max_length': Maximum length successfully encoded
            - 'recommended_token_size': Recommended token budget (90% of max)
            - 'trials_per_length': Dictionary of successful trials per length
            - 'device': Device used for testing

    Raises:
        ValueError: If encoder doesn't have required attributes or torch not available
    """
    if torch is None:
        raise ValueError("PyTorch is required for token size estimation")

    if not hasattr(encoder, "encode") or not hasattr(encoder, "device"):
        raise ValueError("encoder must be a ProstT5ThreeDiEncoder instance")

    if not hasattr(encoder, "_load_model"):
        raise ValueError("encoder must have _load_model method")

    logger.info("Starting token size estimation on device: %s", encoder.device)
    logger.info("Testing range: %d to %d (step: %d)", start_length, end_length, step)

    # Load the model before starting estimation
    logger.info("Loading model...")
    encoder._load_model()

    max_successful_length = 0
    trials_per_length: Dict[int, int] = {}

    for total_length in range(start_length, end_length + 1, step):
        logger.info("Testing total length: %d amino acids", total_length)

        successful_trials = 0

        for trial in range(num_trials):
            try:
                # Generate proteins that combine to target length
                proteins = generate_combined_proteins(
                    total_length,
                    base_length=base_protein_length,
                    seed=trial,  # Different seed per trial
                )

                logger.info(
                    "  Trial %d/%d: encoding %d proteins (total %d AA)",
                    trial + 1,
                    num_trials,
                    len(proteins),
                    sum(len(p) for p in proteins),
                )

                # Attempt encoding with token budget
                # Use the encoder's token_budget_batches for realistic batching
                batches = list(encoder.token_budget_batches(proteins, total_length))

                # Try to encode all batches
                for batch_idx, batch in enumerate(batches):
                    batch_seqs = [item.seq for item in batch]
                    _ = encoder._encode_batch(batch_seqs)
                    logger.info(
                        "    Batch %d/%d encoded successfully",
                        batch_idx + 1,
                        len(batches),
                    )

                successful_trials += 1
                logger.info("  Trial %d/%d: SUCCESS", trial + 1, num_trials)

            except torch.cuda.OutOfMemoryError as e:
                logger.warning(
                    "  Trial %d/%d: Out of memory at length %d: %s",
                    trial + 1,
                    num_trials,
                    total_length,
                    str(e),
                )
                # Clear cache and break on OOM
                if torch.cuda.is_available():
                    torch.cuda.empty_cache()
                break
            except Exception as e:
                logger.error(
                    "  Trial %d/%d: Unexpected error at length %d: %s",
                    trial + 1,
                    num_trials,
                    total_length,
                    str(e),
                )
                # Don't break on other errors, might be transient
                continue

        trials_per_length[total_length] = successful_trials

        # If no trials succeeded, we've hit the limit
        if successful_trials == 0:
            logger.info(
                "No successful trials at length %d, stopping estimation", total_length
            )
            break

        # Update max successful length if at least one trial succeeded
        if successful_trials > 0:
            max_successful_length = total_length

    # Calculate recommended token size (90% of max for safety margin)
    recommended = (
        int(max_successful_length * 0.9) if max_successful_length > 0 else start_length
    )

    results: Dict[str, Any] = {
        "max_length": max_successful_length,
        "recommended_token_size": recommended,
        "trials_per_length": trials_per_length,
        "device": encoder.device,
    }

    logger.info("=" * 60)
    logger.info("Token Size Estimation Complete")
    logger.info("=" * 60)
    logger.info("Device: %s", results["device"])
    logger.info("Max successful length: %d amino acids", results["max_length"])
    logger.info(
        "Recommended token size: %d amino acids", results["recommended_token_size"]
    )
    logger.info("Trials per length: %s", results["trials_per_length"])
    logger.info("=" * 60)

    return results



"""CLI smoke tests to verify commands run without crashing."""

import re

import pytest

# Skip all CLI tests if typer is not installed
pytest.importorskip("typer")

from typer.testing import CliRunner

from genome_entropy.cli.main import app

runner = CliRunner()


def test_cli_help() -> None:
    """Test that main help command works."""
    result = runner.invoke(app, ["--help"])
    assert result.exit_code == 0
    assert "genome_entropy" in result.stdout.lower() or "DNA to 3Di" in result.stdout


def test_cli_version() -> None:
    """Test version flag."""
    result = runner.invoke(app, ["--version"])
    assert result.exit_code == 0
    assert "version" in result.stdout.lower() or "0.1.2" in result.stdout


def test_download_command_help() -> None:
    """Test download command help."""
    result = runner.invoke(app, ["download", "--help"])
    assert result.exit_code == 0
    assert "model" in result.stdout.lower()


def test_orf_command_help() -> None:
    """Test orf command help."""
    result = runner.invoke(app, ["orf", "--help"])
    assert result.exit_code == 0
    assert "input" in result.stdout.lower()
    assert "output" in result.stdout.lower()


def test_translate_command_help() -> None:
    """Test translate command help."""
    result = runner.invoke(app, ["translate", "--help"])
    assert result.exit_code == 0
    assert "input" in result.stdout.lower()
    assert "output" in result.stdout.lower()


def test_encode3di_command_help() -> None:
    """Test encode3di command help."""
    result = runner.invoke(app, ["encode3di", "--help"])
    assert result.exit_code == 0
    assert "input" in result.stdout.lower()
    assert "output" in result.stdout.lower()


def test_entropy_command_help() -> None:
    """Test entropy command help."""
    result = runner.invoke(app, ["entropy", "--help"])
    assert result.exit_code == 0
    assert "input" in result.stdout.lower()
    assert "output" in result.stdout.lower()


def test_run_command_help() -> None:
    """Test run command help."""
    result = runner.invoke(app, ["run", "--help"])
    assert result.exit_code == 0
    assert "input" in result.stdout.lower()
    assert "output" in result.stdout.lower()


def test_run_command_has_new_options() -> None:
    """Test that run command has encoding-size, log-level, and log-file options."""
    result = runner.invoke(app, ["run", "--help"])
    assert result.exit_code == 0
    # Remove ANSI codes to check for options
    clean_output = re.sub(r"\x1b\[[0-9;]*m", "", result.stdout)
    assert "--encoding-size" in clean_output
    assert "--log-level" in clean_output
    assert "--log-file" in clean_output


def test_run_command_encoding_size_option() -> None:
    """Test that run command accepts --encoding-size option."""
    result = runner.invoke(app, ["run", "--help"])
    assert result.exit_code == 0
    # Remove ANSI codes and check
    clean_output = re.sub(r"\x1b\[[0-9;]*m", "", result.stdout)
    assert "--encoding-size" in clean_output
    # Make sure it describes encoding size
    assert (
        "encoding size" in clean_output.lower() or "amino acids" in clean_output.lower()
    )


def test_run_command_log_level_option() -> None:
    """Test that run command accepts --log-level option."""
    result = runner.invoke(app, ["run", "--help"])
    assert result.exit_code == 0
    # Remove ANSI codes and check
    clean_output = re.sub(r"\x1b\[[0-9;]*m", "", result.stdout)
    assert "--log-level" in clean_output
    # Make sure it describes logging level
    assert (
        "logging level" in clean_output.lower() or "log level" in clean_output.lower()
    )


def test_cli_no_command() -> None:
    """Test CLI with no command shows help."""
    result = runner.invoke(app, [])
    # Should show help and exit gracefully
    assert "genome_entropy" in result.stdout.lower() or "DNA" in result.stdout


def test_cli_invalid_command() -> None:
    """Test CLI with invalid command."""
    result = runner.invoke(app, ["invalid_command"])
    assert result.exit_code != 0


def test_estimate_tokens_command_help() -> None:
    """Test estimate-tokens command help."""
    result = runner.invoke(app, ["estimate-tokens", "--help"])
    assert result.exit_code == 0
    assert "token size" in result.stdout.lower()
    assert "model" in result.stdout.lower()
    assert "device" in result.stdout.lower()
    # Remove ANSI codes to check for options
    clean_output = re.sub(r"\x1b\[[0-9;]*m", "", result.stdout)
    assert "--log-level" in clean_output
    assert "--log-file" in clean_output


def test_estimate_tokens_log_level_option() -> None:
    """Test that estimate-tokens accepts --log-level option."""
    # Just checking help to ensure the option is recognized
    result = runner.invoke(app, ["estimate-tokens", "--help"])
    assert result.exit_code == 0
    # Remove ANSI codes and check
    clean_output = re.sub(r"\x1b\[[0-9;]*m", "", result.stdout)
    assert "--log-level" in clean_output
    # Make sure it describes logging level
    assert (
        "logging level" in clean_output.lower() or "log level" in clean_output.lower()
    )



"""ORF extraction command."""

from pathlib import Path

try:
    import typer
except ImportError:
    typer = None


def orf_command(
    input: Path = typer.Option(
        ...,
        "--input",
        "-i",
        help="Input FASTA file",
        exists=True,
        dir_okay=False,
    ),
    output: Path = typer.Option(
        ...,
        "--output",
        "-o",
        help="Output JSON file with ORF records",
    ),
    table: int = typer.Option(
        11,
        "--table",
        "-t",
        help="NCBI genetic code table ID",
    ),
    min_nt: int = typer.Option(
        90,
        "--min-nt",
        help="Minimum ORF length in nucleotides",
    ),
) -> None:
    """Extract ORFs from DNA sequences.
    
    Finds all Open Reading Frames (ORFs) in input FASTA sequences
    using the get_orfs binary. Outputs ORF records as JSON.
    """
    try:
        from ...io.fasta import read_fasta
        from ...io.jsonio import write_json
        from ...orf.finder import find_orfs
        
        typer.echo(f"Reading FASTA from: {input}")
        sequences = read_fasta(input)
        typer.echo(f"  Found {len(sequences)} sequence(s)")
        
        typer.echo(f"\nFinding ORFs (min length: {min_nt} nt, table: {table})...")
        orfs = find_orfs(sequences, table_id=table, min_nt_length=min_nt)
        typer.echo(f"  Found {len(orfs)} ORF(s)")
        
        typer.echo(f"\nWriting results to: {output}")
        write_json(orfs, output)
        
        typer.echo("✓ ORF extraction complete!")
        
    except Exception as e:
        typer.echo(f"Error: {e}", err=True)
        raise typer.Exit(3)
